/**
 * rxrunscript
 * Run shell script in child process, Output Observable<Buffer>
 *
 * @version 6.1.0
 * @author waiting
 * @license MIT
 * @link https://github.com/waitingsong/rxrunscript#readme
 */

import { spawn } from 'child_process';
import { fromEvent, NEVER, race, EMPTY, of, merge, throwError } from 'rxjs';
import { take, takeUntil, shareReplay, buffer, filter, map, skipUntil, mergeMap, ignoreElements, finalize, tap, defaultIfEmpty, catchError } from 'rxjs/operators';
import { join } from '@waiting/shared-core';

function bindProcClose(proc) {
    const stream$ = fromEvent(proc, 'close').pipe(
    // tap(val => console.log('bindProcClsoe data:', val)),
    take(1));
    return stream$;
}

function bindProcError(proc, closingNotifier$) {
    const stream$ = fromEvent(proc, 'error').pipe(take(1), 
    // tap(err => {
    //   throw err
    // }),
    takeUntil(closingNotifier$));
    return stream$;
}

function bindProcExit(proc) {
    const stream$ = fromEvent(proc, 'exit').pipe(
    // tap(val => console.log('bindProcExit data:', val)),
    take(1));
    return stream$;
}

function bindStderrData(stderr, takeUntilNotifier$, skipUntilNofifier$, bufMaxSize) {
    if (!stderr) {
        throw new Error('stderr null');
    }
    const take$ = takeUntilNotifier$.pipe(take(1), shareReplay());
    const skip$ = skipUntilNofifier$.pipe(take(1), shareReplay(1));
    const event$ = bufMaxSize > 0
        ? fromEvent(stderr, 'data')
        : NEVER;
    const data$ = event$.pipe(takeUntil(take$), take(bufMaxSize), 
    // tap(buf => console.info('inner stderr1', buf.toString(), buf.byteLength)),
    buffer(take$), 
    // buffer() may emit blank data, so filter
    filter(arr => arr.length > 0), 
    // map(Buffer.concat), // !! works not output empty array
    map(arr => Buffer.concat(arr)));
    const ret$ = data$.pipe(skipUntil(skip$));
    return ret$;
}

function bindStdinData(stdin, inputData$) {
    if (!stdin) {
        throw new Error('stdint null');
    }
    const input$ = inputData$.pipe(mergeMap((data) => {
        // console.log('bindStdinData:', data)
        // debug in vsc below will cause EPIPE error
        return new Promise((done, reject) => {
            const ended = stdin.writableEnded;
            // prevent "Uncaught Error [ERR_STREAM_WRITE_AFTER_END]: write after end",
            // even if use try/catch
            if (ended) {
                const err = new Error(' write after end');
                // @ts-expect-error
                err.code = 'ERR_STREAM_WRITE_AFTER_END';
                return reject(err);
            }
            stdin.write(data, (err) => {
                if (err) {
                    reject(err);
                    return;
                }
                done(data);
            });
        });
    }), ignoreElements(), finalize(() => {
        stdin.end();
    }));
    return input$;
}

function bindStdoutData(stdout, takeUntilNotifier$) {
    if (!stdout) {
        throw new Error('stdout null');
    }
    const take$ = takeUntilNotifier$.pipe(take(1), shareReplay());
    const data$ = fromEvent(stdout, 'data').pipe(
    // tap(buf => console.info('stdout:', buf.toString())),
    takeUntil(take$));
    return data$;
}

function bindEvent(proc, stderrMaxBufferSize, msgPrefixOpts, script, // for throw error
inputStream) {
    const { stderrPrefix } = msgPrefixOpts;
    const close$ = bindProcClose(proc);
    const exit$ = bindProcExit(proc);
    const closeOrExit$ = race(close$, exit$).pipe(
    // tap(codeSignal => {
    //   console.log('close or exit event', codeSignal)
    // }),
    shareReplay(1));
    const stdout$ = bindStdoutData(proc.stdout, closeOrExit$);
    const error$ = bindProcError(proc, closeOrExit$);
    const skipUntilNotifier$ = closeOrExit$.pipe(mergeMap(([code]) => {
        return code === 0 || code === null ? EMPTY : of(void 0);
    }));
    const stderr$ = bindStderrData(proc.stderr, closeOrExit$, skipUntilNotifier$, stderrMaxBufferSize).pipe(tap((buf) => {
        const msg = buf.toString();
        throw new Error(`${stderrPrefix} ${script}\n${msg}`);
    }));
    const stdin$ = inputStream && typeof inputStream.subscribe === 'function'
        ? bindStdinData(proc.stdin, inputStream)
        : EMPTY;
    const ret$ = merge(stdout$, stderr$, stdin$, error$).pipe(finalize(() => {
        // sub process may not stop
        // link: https://nodejs.org/api/child_process.html#child_process_subprocess_kill_signal
        proc.killed || proc.kill();
    }), defaultIfEmpty(Buffer.from('')));
    return ret$;
}

const initialSpawnOpts = {
    cwd: process.cwd(),
    env: { ...process.env },
    stdio: 'pipe',
};
const initialMsgPrefixOpts = {
    errPrefix: 'Run script with error:',
    stderrPrefix: 'Run script with stderr:',
};
const initialRxRunOpts = {
    maxCmdLength: 2048,
    msgPrefixOpts: initialMsgPrefixOpts,
    stderrMaxBufferSize: 200,
};

function processOpts(options) {
    const { args, initialRxRunOpts } = options;
    const spawnOpts = processSpawnOpts(initialRxRunOpts, options.spawnOpts);
    const cmd = processCommand(options.command, initialRxRunOpts.msgPrefixOpts.errPrefix, spawnOpts);
    // eslint-disable-next-line @typescript-eslint/prefer-optional-chain
    const runArgs = args && args.length ? [...args] : [];
    const errScript = runArgs.length
        ? `"${cmd} ${runArgs.join(' ')}"`
        : `"${cmd}"`;
    return {
        command: cmd,
        args: runArgs,
        spawnOpts,
        errScript,
    };
}
function processSpawnOpts(defaultRxrunOpts, options) {
    const opts = (options ? { ...initialSpawnOpts, ...options } : initialSpawnOpts);
    opts.shell = true;
    /* istanbul ignore next */
    if (process.platform === 'win32') {
        opts.windowsVerbatimArguments = true;
    }
    /* istanbul ignore else */
    if (typeof opts.stderrMaxBufferSize !== 'number' || opts.stderrMaxBufferSize < 0) {
        opts.stderrMaxBufferSize = +defaultRxrunOpts.stderrMaxBufferSize;
    }
    opts.msgPrefixOpts = { ...initialMsgPrefixOpts };
    return opts;
}
function processCommand(command, errPrefix, spawnOpts) {
    let cmd = command ? command.trim() : '';
    /* istanbul ignore else */
    if (!cmd) {
        throw new TypeError(`${errPrefix}\nRun command is blank`);
    }
    const cmdLead = cmd.slice(0, 3).replace(/\\/ug, '/');
    /* istanbul ignore else */
    if (spawnOpts.cwd && (cmdLead === '../' || cmdLead.startsWith('./'))) {
        cmd = join(spawnOpts.cwd, cmd);
    }
    /* istanbul ignore else */
    if (typeof spawnOpts.maxCmdLength === 'number'
        && spawnOpts.maxCmdLength > 0
        && cmd.length > spawnOpts.maxCmdLength) {
        throw new TypeError(`${errPrefix}\nCommand length exceed ${spawnOpts.maxCmdLength}\n` + cmd.slice(1024) + ' ...');
    }
    return cmd;
}

/**
  * Reactive Running of Command Script
  * @param command {string}
  * @param args {string[]|null}
  * @param options {Partial<RxRunOpts>|null}
  *
  * @returns Observable<Buffer>
  * @link https://nodejs.org/api/child_process.html#child_process_child_process_spawn_command_args_options
  */
function run(command, args, options) {
    const opts = processOpts({
        command, args, spawnOpts: options, initialRxRunOpts,
    });
    const { errScript } = opts;
    const { errPrefix, stderrPrefix } = opts.spawnOpts.msgPrefixOpts;
    const { inputStream } = opts.spawnOpts;
    const proc$ = runSpawn(opts.command, opts.args, opts.spawnOpts);
    const ret$ = proc$.pipe(mergeMap((proc) => {
        return bindEvent(proc, opts.spawnOpts.stderrMaxBufferSize, opts.spawnOpts.msgPrefixOpts, opts.errScript, inputStream);
    }), catchError((err) => {
        const msg = err.message;
        if (!msg.startsWith(errPrefix) && !msg.startsWith(stderrPrefix)) {
            err.message = `${errPrefix} ${errScript}\n` + msg;
        }
        throw err;
    }));
    return ret$;
}
function runSpawn(command, runArgs, spawnOpts) {
    try {
        const proc = spawn(command, runArgs, spawnOpts);
        return of(proc);
    }
    catch (ex) {
        return throwError(() => ex);
    }
}
function escapeShell(cmd) {
    return cmd.replace(/(["\s'$`\\])/ug, '\\$1');
}

export { escapeShell, run };
//# sourceMappingURL=rxrunscript.esm.js.map
